{-
    Main Browser.
    Copyright (C) 2007 Luis Francisco Araujo <araujo@gentoo.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
-}

module Browser
    where

import Emerge
import Menus
import Util
import Web
import Render
import Graphics.UI.Gtk
import System.Directory
import Control.Monad
import Control.Concurrent
import Data.List
import Data.Char (isAlpha)
import Data.Maybe
import System.Exit
import qualified Control.Exception as E 
import System.IO
import System.Process
import Graphics.UI.Gtk.MozEmbed

-- | Browser type.
type Browser = (VBox, Notebook, TextView, Statusbar, 
		ProgressBar, MozEmbed, Tooltips)

buttonIconList :: [(String, String)]
buttonIconList = 
    [ ("queue", "Calculate the dependencies for a selected package and add them into the package queue")
    , ("revdep", "Calculate the direct reverse dependencies for a selected package and add them into the package queue")
    , ("add", "Add selected package to the processing queue")
    , ("remove", "Remove a selected package from the queue")
    , ("clear", "Clear the main package queue")
    , ("world", "Calculate the world dependencies and add them into the main package queue")
    , ("neworld", "Calculate the world dependencies with the 'newuse' flag enabled and add them into the main package queue")
    , ("sync", "Synchronize the portage tree with the remote repositories (sync)")
    , ("emerge", "Install the packages listed on the queue")
    , ("unmerge", "Uninstall the packages listed on the queue")
    , ("binary", "Install and make binary tarballs from the package listed on the queue")
    , ("usepkg", "Install and use binary tarballs (if available) from the packages listed on the queue")
    , ("fetch", "Only fetch the packages listed on the queue. Do not install them")
    , ("system", "Caculate the 'system' packages and add them into the queue")
    , ("all", "List all packages installed in the system into the queue")
    ]

cquestions :: [String]
cquestions = [ "Do you want to install queded packages?"
	     , "Do you want to un-install queded packages?" 
	     , "Do you want to install and build binary tarballs for queded packages?"
	     , "Do you want to install queded packages?\n(using available binary tarballs)"
	     , "Do you only want to fetch queded packages?" ]

portageBrowser :: FilePath -> IO Browser
portageBrowser prepo = 
    do
    timeoutAddFull (yield >> return True) priorityDefaultIdle 50
    let portagerepo = prepo ++ "/"
    -- This is the main vertical box containing
    -- all the widgets.
    mainvbox <- vBoxNew False 3
    browservbox <- vBoxNew False 3
    -- Logo.
    image <- imageNewFromFile windowlogo
    -- Let's use ToolBars for the buttons.
    buttonstoolbar <- toolbarNew
    tooltips <- tooltipsNew
    toolbarSetTooltips buttonstoolbar True
    [ queuebutton , revdepbutton,  addqueuebutton, removequeuebutton
      , clearqueuebutton, worldbutton, neworldbutton, syncbutton
      , emergebutton, unmergebutton, binarybutton, usepkgbutton
      , fetchbutton, systembutton, allbutton ]
         <- mapM (toolButton tooltips buttonstoolbar) buttonIconList
    -- eix entry text search.
    checkbox <- checkButtonNew
    searchbox <- comboBoxEntryNewText
    searchlabel <- imageNewFromFile searchicon
    searchbutton <- buttonNew
    containerAdd searchbutton searchlabel
    -- Tell what some of these options do.
    tooltipsSetTip tooltips checkbox "Search package matching exact name" []
    tooltipsSetTip tooltips searchbutton "Search package" []
    labelsearchbox <- hBoxNew False 5
    opt <- labelNew Nothing
    labelSetMarkup opt ("<i><u>" ++ portagerepo ++ "</u></i>")
    boxPackStart labelsearchbox image PackNatural 0
    boxPackStart labelsearchbox opt PackNatural 0
    boxPackEnd labelsearchbox checkbox PackNatural 0
    boxPackEnd labelsearchbox searchbutton PackNatural 0
    boxPackEnd labelsearchbox searchbox PackNatural 0
    boxPackStart browservbox labelsearchbox PackNatural 0
    -- Add the main toolbar with toolbuttons to the window.
    boxPackStart browservbox buttonstoolbar PackNatural 5
    {- | Start adding the Portage Tree Views.
       Get the portage tree root contents. -}
    portagetree <- (getDirectory portagerepo >>= return . filterDirs removeddirs . removeOddDirs)
    -- initialize all three liststore.
    (categorytv, categoryscroll, _) <- createStore "Category" SelectionSingle False
				       (Just (categoryicon, portagetree))
    (packagetv, packagescroll, _) <- createStore "Package" SelectionSingle False Nothing
    (versiontv, versionscroll, _) <- createStore "Version" SelectionSingle False Nothing
    -- Portage tree view main box.
    middlepanel <- hBoxNew False 0
    mapM_ (flip (flip (boxPackStart middlepanel) PackGrow) 3)
	      [categoryscroll, packagescroll, versionscroll]
    -- Add the main three treeviews to the window.
    -- Create a frame around them.
    treeframe <- makeFrame "Portage Browser" 0.50 0.50
    containerSetBorderWidth middlepanel 5
    containerAdd treeframe middlepanel
    boxPackStart browservbox treeframe PackGrow 0
    -- Let's use paneds to allow widgets to be resized.
    -- Vertical paned.
    vpaned <- vPanedNew
    -- Horizontal paned.
    hpaned <- hPanedNew
    -- Makes the browser box resizable.
    panedPack1 vpaned browservbox True True
    -- Start creating the notebook panel.
    processpanel <- notebookNew
    notebookSetScrollable processpanel True
    notebookSetPopup processpanel True
    -- Different boxes containing the information windows
    -- for the notebook pages.
    (mozframe, mozwindow) <- webBrowser
    -- Load the initial documentation.
    mozEmbedLoadUrl mozwindow hdoc
    -- Package mozilla browser.
    panedPack1 hpaned mozframe True True
    (logscroll, logview) <- makeView False
    (ebuildscroll, ebuildview) <- makeView True
    (clogscroll, clogview) <- makeView False
    (pkgqueuetv, pkgqueuescroll) <- createQueueStore
    -- Set the name of this widget to make it accessible to the
    -- 'queue' operations.
    widgetSetName pkgqueuescroll "scrollview"
    widgetSetName pkgqueuetv "treeview"
    -- Notebook pages.
    -- Log view.
    notebookAppendPage processpanel logscroll "Log"
    -- Queue package window.
    notebookAppendPage processpanel pkgqueuescroll "Main Queue"
    -- Ebuild text window.
    notebookAppendPage processpanel ebuildscroll "Ebuild"
    -- ChangeLog text window.
    notebookAppendPage processpanel clogscroll "Changelog"
    -- Add the notebook panel into a frame
    controlframe <- makeFrame "Control Panel" 0.0 0.50
    containerSetBorderWidth processpanel 5
    containerAdd controlframe processpanel
    -- Add the controlframe into the horizontal paned.
    panedPack2 hpaned controlframe True False
    -- Packs the horizontal paned inside the vertical one.
    -- This allows resizing both the browser box and the info/mozilla widgets.
    panedPack2 vpaned hpaned True False
    -- Add the vertical paned to the main box.
    boxPackStart mainvbox vpaned PackGrow 0
    -- Build the status display box.
    statusbox <- hBoxNew False 0
    progbar <- progressBarNew
    statbar <- statusbarNew
    updateStatBar statbar "Welcome to himerge!"
    boxPackStart statusbox statbar PackGrow 0
    boxPackEnd statusbox progbar PackNatural 0
    -- Add the last box into the main box.
    boxPackStart mainvbox statusbox PackNatural 0
    -- Search the package using eix.
    searchbutton `onClicked` comboBoxSearchEntry searchbox
		     (search portagerepo packagetv categorytv versiontv statbar checkbox mozwindow)
    searchbox `onKeyRelease` comboBoxSearchEntryEvent searchbox
		  (search portagerepo packagetv categorytv versiontv statbar checkbox mozwindow)
    -- Update the category view.
    categorytv `onCursorChanged` 
	       updateCategory packageicon portagerepo categorytv packagetv
    -- Update the package version view and information mozilla browser.
    packagetv `onCursorChanged`
		  (updatePackage ebuildicon portagerepo packagetv versiontv >>=
		   (\ pkgname -> do
		    (Just tvc) <- treeViewGetColumn packagetv 0
		    (Just cat) <- treeViewColumnGetTitle tvc
		    updateMozilla mozwindow (" -e " ++ cat ++ "/" ++ pkgname)))
    -- Update package specific information views.
    versiontv `onCursorChanged` updateInfo portagerepo versiontv (ebuildview, clogview) statbar
    -- Package popup menu.
    versiontv `onButtonRelease` packagePopupMenu portagerepo versiontv statbar
    -- Queue operations.
    queuebutton `onToolButtonClicked` pkgDependency versiontv processpanel logview statbar progbar
    revdepbutton `onToolButtonClicked` pkgReverseDependency versiontv processpanel logview statbar progbar
    addqueuebutton `onToolButtonClicked` addAtQueue versiontv processpanel statbar
    removequeuebutton `onToolButtonClicked` removePageElement processpanel
    clearqueuebutton `onToolButtonClicked` clearQueue processpanel statbar
    worldbutton `onToolButtonClicked` 
		(writeLog logview msgworld >>
		 emergeQueue "--deep --update world" msgworld processpanel logview statbar progbar)
    neworldbutton `onToolButtonClicked`
		  (writeLog logview msgworldnewuse >>
		   emergeQueue "--newuse --deep --update world" msgworldnewuse processpanel logview statbar progbar)
    systembutton `onToolButtonClicked`
		 (writeLog logview msgsystem >>
		  emergeQueue "--newuse --deep --update system" msgsystem processpanel logview statbar progbar)
    allbutton `onToolButtonClicked` 
		  (writeLog logview msginstalled >> pkgInstalled processpanel logview msginstalled statbar progbar)
    -- Main emerge operations.
    syncbutton `onToolButtonClicked` sync processpanel logview statbar progbar
    emergebutton `onToolButtonClicked`
		 popSelectWindow (cquestions !! 0) (emergePackages processpanel logview statbar progbar)
    unmergebutton `onToolButtonClicked`
		  popSelectWindow (cquestions !! 1) (unmergePackages processpanel logview statbar progbar)
    binarybutton `onToolButtonClicked` 
		 popSelectWindow (cquestions !! 2) (binaryPackages processpanel logview statbar progbar)
    usepkgbutton `onToolButtonClicked` 
		 popSelectWindow (cquestions !! 3) (useBinaryPackages processpanel logview statbar progbar)
    fetchbutton `onToolButtonClicked`
		popSelectWindow (cquestions !! 4) (fetchPackages processpanel logview statbar progbar)
    -- return this main vbox.
    return (mainvbox, processpanel, logview, statbar, progbar, mozwindow, tooltips)
	where
	msgworld = "Calculating world dependencies."
	msgworldnewuse = "Calculating world dependencies (newuse)."
	msgsystem = "Calculating system dependencies."
	msginstalled = "Calculating installed packages."

getSelectedPkg :: TreeView -> IO (Maybe String)
{- | Get the package name from a tree view (fromtv). -}
getSelectedPkg fromtv =
    do
    [(fromstore, sel)] <- storeAndSelector [fromtv] 
    maybeiter <- treeSelectionGetSelected sel
    case maybeiter of
      Nothing -> return Nothing
      (Just node) -> getCategoryPackage fromtv fromstore node >>=
		     \s -> return $ Just s

runForPkgQueue :: TreeView -> Notebook -> TextView
	       -> (String -> String -> Notebook -> TextView -> Statusbar -> ProgressBar -> IO ())
	       -> Statusbar -> ProgressBar -> String -> IO ()
runForPkgQueue fromtv panel logview func statbar progbar extramsg =
    do
    selectedpkg <- getSelectedPkg fromtv
    case selectedpkg of
      Nothing -> updateStatBar statbar "No package version selected."
      (Just pkg) -> -- =cat/pkg-ver
		     do
		     let loginfo = ("Calculating" ++ extramsg ++ " dependencies for " ++ pkg)
		     -- Write operation log.
		     writeLog logview loginfo
		     func ("=" ++ pkg) loginfo panel logview statbar progbar

makeQueueFromCommand :: String -> String -> (Handle -> IO [String]) -> String 
		     -> Notebook -> TextView -> Statusbar -> ProgressBar -> String -> IO ()
{- | Apply a command over a package name and gets its output into
   the package Queue. -}
makeQueueFromCommand command statmsg func pkgtoken panel logview statbar progbar extramsg =
    do
    (_,out,err,ph) <- runInteractiveCommand command
    barthread <- forkIO $ updateBar progbar
    nout <- E.evaluate out
    statid <- forkIO $ refreshStatBar statbar statmsg
    forkIO $ onExit ph ((.) ((>>) (killThread statid)) (handleExitCode nout err barthread))
    return ()
    where
    handleExitCode handle errhandle barthread exitcode =
	killThread barthread >>
	case exitcode of
	  ExitSuccess -> do
		      let pkg = last $ words pkgtoken
		      pkglist <- func handle
                      case pkglist of
		         [] -> do
			       let rlog = ("No valid dependencies for " ++ pkg)
			       writeLog logview rlog
			       updateStatBar statbar rlog
			       progressBarSetFraction progbar 0.0
			 pkglist' -> do
			             -- Make a new Queue for each process.
				     (pkgqueuetv, pkgqueuescroll) <- createQueueStore
				     -- Name this widget to keep track of the correct package values.
				     widgetSetName pkgqueuetv "treeview"
				     widgetSetName pkgqueuescroll "scrollview"
				     -- Build the panel page.
				     (tooltips, _, (Just closebutton))
					 <- buildPanelTab pkgqueuescroll panel ("Queue: " ++ pkgtoken) CloseButton
				     widgetShowAll pkgqueuescroll
				     notebookSetCurrentPage panel (-1)
				     -- Connect this button to the close signal.
				     closebutton `onClicked`	    
						     (popSelectWindow "Do you really want to close this tab?"
                 				      (tooltipsEnable tooltips >> closeTab panel pkgqueuescroll))
				     depst <- getStoreFromView pkgqueuetv
				     listStoreClear depst
				     -- Add packages into the queue.
				     pixbuf <- pixbufNewFromFile ebuildicon
				     storeQueueObjects depst pixbuf pkglist'
				     let rlog = (pkg ++ extramsg ++ " dependencies successfully calculated.")
				     writeLog logview rlog
				     updateStatBar statbar rlog
				     progressBarSetFraction progbar 1.0
	  _ -> do
	       hGetContents errhandle >>= hPutStr stderr
	       let rlog = ("Error calculating package" ++ extramsg ++ " dependencies.")
	       writeLog logview rlog
	       emergeError rlog handle
	       updateStatBar statbar rlog
	       progressBarSetFraction progbar 0.0

emergeQueue :: String -> String -> Notebook -> TextView
	    -> Statusbar -> ProgressBar -> IO ()
emergeQueue pkgtoken statmsg panel logview statbar progbar =
    makeQueueFromCommand (emergepath ++ " --verbose --pretend " ++ pkgtoken) 
                         statmsg emergeDeps pkgtoken panel logview statbar progbar []

pkgDependency :: TreeView -> Notebook -> TextView 
	      -> Statusbar -> ProgressBar -> IO ()
{- | Calculate the dependency queue of a package. -}
pkgDependency fromtv panel logview statbar progbar =
    runForPkgQueue fromtv panel logview emergeQueue statbar progbar []

pkgReverseDependency :: TreeView -> Notebook -> TextView 
		     -> Statusbar -> ProgressBar -> IO ()
{- | Calculte the reverse dependency queue of a package. -}
pkgReverseDependency fromtv panel logview statbar progbar =
    runForPkgQueue fromtv panel logview func statbar progbar " reverse"
	where
	func pkgtoken statmsg panel' logview' statbar' progbar' =
	    makeQueueFromCommand (equerypath ++ "depends " ++ pkgtoken) statmsg
				 equeryOutput pkgtoken panel' logview' statbar' progbar' " reverse"

pkgInstalled :: Notebook -> TextView -> String -> Statusbar -> ProgressBar -> IO ()
pkgInstalled panel logview statmsg statbar progbar =
    makeQueueFromCommand (qlistpath ++ "--verbose --nocolor --installed") statmsg
                         equeryOutput "all-packages" panel logview statbar progbar []

addAtQueue :: TreeView -> Notebook -> Statusbar -> IO ()
addAtQueue fromtv panel statbar =
    isCurrentPagePackagePanel panel func
    where
    func totv =
	do
	[(fromstore, sel), (tostore,_)] <- storeAndSelector [fromtv, totv]
	maybeiter <- treeSelectionGetSelected sel
	case maybeiter of
	   Nothing -> updateStatBar statbar "No package version selected."
	   (Just node) -> do
      	                  catpkg <- getCategoryPackage fromtv fromstore node
			  updateStatBar statbar (catpkg ++ " added to queue.")
			  pixbuf <- pixbufNewFromFile ebuildicon
			  storeQueueObjects tostore pixbuf [catpkg]

clearQueue :: Notebook -> Statusbar -> IO ()
{- | Removed all the packages in the queue if any. -}
clearQueue panel statbar =
    isCurrentPagePackagePanel panel
	     (\ pkgtv -> getStoreFromView pkgtv >>= listStoreClear >> 
	      updateStatBar statbar "Cleared package queue.")

updateCategory :: FilePath -> FilePath -> TreeView 
	       -> TreeView -> IO ()
{- | Update the packages views with the category
   information available. -}
updateCategory image repopath tv newtv =
    do
    [(store, sel), (newstore, _)] <- storeAndSelector [tv, newtv]
    rows <- treeSelectionGetSelectedRows sel
    case rows of
     [[node]] -> do
	     (GVstring (Just catname)) <- valueWithIterTreeModel store node textColumn
	     dir <- getDirectory (repopath ++ catname ++ "/")
	     initStoringObj newtv newstore catname image dir
	     iter <- treeModelGetIter newstore [0]
	     if isJust iter 
		then scrollViewToCell newstore (fromJust iter) newtv
		else popErrorWindow "No valid category."
     _ -> return ()

updatePackage :: FilePath -> FilePath -> TreeView -> TreeView -> IO String
{- | Update the package versions views with the package
   information available. -}
updatePackage image repopath tv newtv =
    do
    [(store, sel), (newstore, _)] <- storeAndSelector [tv, newtv]
    rows <- treeSelectionGetSelectedRows sel
    case rows of
     [[node]] -> do
		 (GVstring (Just pkgname)) <- valueWithIterTreeModel store node textColumn
		 (Just tvc) <- treeViewGetColumn tv 0
		 (Just cat) <- treeViewColumnGetTitle tvc
		 let fullcatpkg = cat ++ "/" ++ pkgname
		 -- Update mozilla browser package information.
		 dir <- getVersion (repopath ++ fullcatpkg ++ "/")
		 initStoringObj newtv newstore fullcatpkg image dir
		 return pkgname
     _ -> return []

updateMozilla :: MozEmbed -> String -> IO ()
updateMozilla mozweb pkgname =
    do
    eix <- eixOutput pkgname
    (tmpfile, tmphandle) <- writeTempFile htmltemp
    hPutStr tmphandle $ render eix
    hFlush tmphandle
    mozEmbedLoadUrl mozweb tmpfile
    removeFile tmpfile `E.catch`
		   \ e -> popErrorWindow $ show e
    hClose tmphandle

updateInfo :: FilePath -> TreeView -> (TextView, TextView) 
	   -> Statusbar -> IO ()
{- | Update the different panel views with the 
   version treeview information available. -}
updateInfo repopath tv (ebuildv, clogv) statbar =
   do 
   [(store, sel)] <- storeAndSelector [tv]
   rows <- treeSelectionGetSelectedRows sel
   case rows of
     [[n]] -> do
	      (Just tvc) <- treeViewGetColumn tv 0
	      (Just cat) <- treeViewColumnGetTitle tvc
	      (GVstring (Just package)) <- valueWithIterTreeModel store n textColumn
              let parsedpath = repopath ++ cat ++ "/"
              let changelogpath = parsedpath ++ "ChangeLog"
	      let ebuildpath = parsedpath ++ package ++ ".ebuild"
	      -- Test for the ebuild (just in case) and Changelog file existence.
              [ebuild, changelog] <- 
		  mapM (\ (file, msg) -> do
			b <- doesFileExist file
			if b then readFile file else return msg)
					 [ (ebuildpath, "No ebuild file!!!.")
					 , (changelogpath, "No Changelog file available.") ] 

	      updateTextBuffer ebuildv ebuild
	      updateTextBuffer clogv changelog
     _ -> updateStatBar statbar "Incorrect option."

initStoringObj :: TreeView -> ListStore -> FilePath -> FilePath 
	       -> [FilePath] -> IO ()
{- Add new objects to a tree store. 
   Used for updating the category and package treeviews. -}
initStoringObj tv newstore pkgorcatname image dir =
    do
    listStoreClear newstore
    (Just tvc) <- treeViewGetColumn tv 0
    treeViewColumnSetTitle tvc pkgorcatname
    pixbuf <- pixbufNewFromFile image
    storeObjects newstore pixbuf dir
    return ()

comboBoxSearchEntryEvent :: ComboBoxEntry -> (String -> String -> IO ())
			 -> Event -> IO Bool
comboBoxSearchEntryEvent sbox func (Key { eventKeyName = "Return" }) =
    comboBoxSearchEntry sbox func >> return True
comboBoxSearchEntryEvent _ _ _ = return False

comboBoxSearchEntry :: ComboBoxEntry -> (String -> String -> IO ()) -> IO ()
{- Entry function for the package search box operation. -}
comboBoxSearchEntry sbox func =
    do
    text <- comboBoxGetActiveText sbox
    case text of
       Nothing -> return ()
       Just [] -> return ()
       Just jtext | null $ words jtext -> return ()
		  | otherwise -> do
				 model <- comboBoxGetModel sbox
				 let textval = stripspaces jtext
				 case model of
				    Nothing -> return ()
				    Just m -> comboboxgetstrings m >>=
					      \ s -> if any (== textval) s
						        then return ()
							else comboBoxPrependText sbox textval
				 out <- eixOutput textval
				 forkIO $ (func textval out >> return ())
				 return ()
	where
	comboboxgetstrings model' = collectIters model' 0 >>= collectStringValues model' 0
	stripspaces = unwords . words

search :: FilePath -> TreeView -> TreeView -> TreeView
       -> Statusbar -> CheckButton -> MozEmbed -> String -> String -> IO ()
search repopath pkgtv cattv vertv statbar checkb mozwindow pkgname eixoutput =
    do
    cb <- toggleButtonGetActive checkb
    let parsed = parseEixOutput pkgname cb eixoutput
    if null (fst parsed)
       then updateStatBar statbar ("Package " ++ pkgname ++ " not found.")
       else searchOn pkgname repopath cattv pkgtv vertv parsed statbar cb mozwindow

parseEixOutput :: String -> Bool -> String -> ([String], [String])
{- Return structure of the form: ([categories], [packages]) -}
parseEixOutput pkgname cbool = findpkg pkgname (if cbool then (==) else findSubstring) [] []
			       . filter (\ c -> if (not . null) c then ((/= ' ') . (!! 0)) c else False)
			       . lines
    where
    findpkg _ _ cats pkgs [] = (nub (cats ++ []), nub (pkgs ++ []))
    findpkg pkgname' func cats pkgs (ps:pss) = 
    -- Break at '/' for "* category/package"
	    if (not . any (== '/')) catpkg
	       then findpkg pkgname' func cats pkgs pss
	       else (if func pkgname' pkg || func pkgname' catpkg
		        then findpkg pkgname' func (cats ++ [cat]) (pkgs ++ [pkg]) pss
		        else findpkg pkgname' func cats pkgs pss)
		   where
		   stripr = takeWhile (/= ' ')
		   stripl = dropWhile (not . isAlpha)
		   catpkg = stripr $ stripl $ ps
		   (cat, (_:pkg)) = break (== '/') catpkg

searchOn :: String -> FilePath -> TreeView -> TreeView -> TreeView 
	 -> ([String], [String]) -> Statusbar -> Bool -> MozEmbed -> IO ()
{- | Search the package on the views. -}
searchOn pkgname repopath cattv pkgtv vertv (cats, pkgs) statbar checkbool mozwindow =
    do
    [(catstore, catsel)] <- storeAndSelector [cattv]
    generalfind findcat 0 catstore catsel "category" (cats, pkgs)
    where
    generalfind _ _ _ _ _ ([], []) = updateStatBar statbar "Package not found."
    generalfind _ _ _ _ _ ([], _) = updateStatBar statbar "Package not found."
    generalfind _ _ _ _ _ (_, []) = searchOn pkgname repopath cattv pkgtv vertv (tail cats, pkgs) 
				             statbar checkbool mozwindow
    generalfind func num store sel label (catss, pkgss) =
	do
	jiter <- treeModelGetIter store [num]
	case jiter of
	   Nothing -> if label == "category" 
		          then generalfind func 0 store sel label (tail catss, pkgss)
			  else generalfind func 0 store sel label (catss, tail pkgss)
	   Just iter ->
	       do
	       (GVstring (Just value)) <- treeModelGetValue store iter textColumn
               let mvalue = if label == "category" then (head catss) else (head pkgss)
	       if (==) value mvalue
		  then func iter sel value store (catss, pkgss) >>
		       -- Exact query for eix if it is enabled the checkbutton.
		       case checkbool of { True -> return " -e " ; False -> return []} >>=
		       \ c -> updateMozilla mozwindow (c ++ pkgname)
		  else generalfind func (num + 1) store sel label (catss, pkgss)
    findcat iter' sel' _ store' xs = 
	do
	treeSelectionSelectIter sel' iter'
	-- Scroll to the category found.
	scrollViewToCell store' iter' cattv
	updateCategory packageicon repopath cattv pkgtv
	[(pkgstore, pkgsel)] <- storeAndSelector [pkgtv]
	generalfind findpkg 0 pkgstore pkgsel "package" xs
    findpkg iter' sel' value' store' _ =
	do
	treeSelectionSelectIter sel' iter'
	-- Scroll to the package found.
	scrollViewToCell store' iter' pkgtv
	updatePackage ebuildicon repopath pkgtv vertv
	updateStatBar statbar ("Package " ++ value' ++ " found.")

scrollViewToCell :: TreeModelClass a => a -> TreeIter -> TreeView -> IO ()
{- Scroll to a specific cell in the view. -}
scrollViewToCell tmodel iter tview =
    do
    treepath <- treeModelGetPath tmodel iter
    (Just tvc) <- treeViewGetColumn tview 0 
    treeViewScrollToCell tview treepath tvc (Just (0,0))